#!/usr/local/bin/perl
#configure_2.2_lvs.pl v0.3 (C) Joseph Mack 1999 mack@wm7d.net
#v0.1 6 Nov 99 for kernel 2.2.12
#v0.2 7 Nov 99 for kernel 2.2.13
#v0.3 14 Nov 99 added option for varying timeout for ppc connnection
#v0.4 17 Nov 99 fixed bug in ppc setup code
#
#
#released under GPL
#This code is part of the Linux Virtual Server (LVS) project
#http://www.linuxvirtualserver.org
#
#Caveat - not all inputs have been tested. 
#Please send bug reports to mack@wm7d.net
#
#
#----------------------------------------------------
#Description:
#
#	Generates rc.d type files for an LVS in which the 
#	director is running a Linux 2.2.x kernel.
#	(the previous version of this file, configure.lvs was for a 
#	director running a Linux 2.0.36 kernel, and was a shell script)
#
#	Input - a conf file which has
#		1. The type of LVS (VS_NAT, VS_TUN or VS_DR)
#		2. The IPs for the director and the servers
#		3. The services being served 
#
#	Outputs
#		1. rc.lvs_xxx file, which is copied to /etc/rc.d on all machines 
#		2. mon_xxx.cf file, which is copied to /etc/mon on the director (if you are using mon) 
#			where xxx is nat, tun or dr
#
#	For information about monitors see the LVS-HOWTO and the 
#		mon site http://www.kernel.org/software/mon/ 
#	The fping monitor will be chosen if there is no monitor for your service.
#	Available monitors:
#		dialin, dns, fping, freespace, ftp, ftp_kernelorg, hpnp, http, http_t, 
#		imap, ldap, msql-mysql, na_quota, netappfree,
#		nntp, ping, pop3, process, rd, reboot, ssh, seq, smtp, tcp, telnet, up_rtt
#	Please read the "Big Caveat" about monitors in the LVS-HOWTO 
#
#----------------------------------------------------------
#
# General setup
use strict;
#v0.4 17 Nov 99 fixed bug in ppc setup code
use Net::DNS;

my ($lvs_type, $connection_parameter);
my $initial_state;

my @monitors = ("dialin", "dns", "fping", "freespace", "ftp", "ftp_kernelorg", "hpnp", "http", "http_t", "imap", "ldap", "msql-mysql", "na_quota", "netappfree", "nntp", "ping", "pop3", "process", "seq", "smtp", "ssh", "rd", "reboot", "tcp", "telnet", "up_rtt");
my $conf_file;
my $director_file = "rc.lvs_${lvs_type}director";
my $server_file = "rc.lvs_${lvs_type}server";
my $rc_file = "rc.lvs_$lvs_type";
my $mon_file = "mon_lvs${lvs_type}.cf";
my @params;
my ($vdevice,$vip, $vnetmask, $vbroadcast); #VIP setup for VS_TUN, VS_DR
my ($inside_device,$inside_ip, $inside_network, $inside_netmask, $inside_broadcast,$dot_quad, $classless, $log_2); #server network for VS_NAT

#service setup
my $default_scheduler = "wlc";
my ($proto, $service, $scheduler, $t_switch, $timeout, $server, $weight, $weight_lvsadm); 
my ($service_name, $port, $protocol, $server_ip, $weight);
my ($server_port, $server_ipport, $server_service_name);

#server setup
my ($slvsdevice, $snetdevice, $server_counter, $monitor); 
my @servers; #list of servers

#other
my $temp;

my $line_counter;
$line_counter=0;
$server_counter=0;

#list of known schedulers (ppc is only supported to 2.2.10)
my @schedulers = ("rr","wrr","lc","wlc","pcc");

#list kernels that this script is thought to work on
my @uname_r_ok = ("2.2.10","2.2.12","2.2.13");

sub calc_log_2
	{
	$log_2 = 0;
	my $temp = $_[0];
	#print "$temp \n";
	while ($temp != 1)
		{
		$temp = $temp >> 1;
		$log_2++;
		#print "$temp $log_2 \n";
		}
	#print "$log_2 \n";
	return ($log_2);
	}

sub dotquad_to_classless
	{
	#takes 255.255.255.0, converts it to 24
	$dot_quad = $_[0];
	#print "dot_quad: $dot_quad \n";
	$dot_quad =~ /(\d{1,3})\.(\d{1,3})\.(\d{1,3})\.(\d{1,3})/ ;
	#print "$1 $2 $3 $4 \n";
	$classless = calc_log_2(eval($1+1)) + calc_log_2(eval($2+1)) + calc_log_2(eval($3+1)) + calc_log_2(eval($4+1)); 
	#print "classless $classless \n";
	return ($classless);
	}

sub parse_service
	{
	my $service = $_[0];	
	my ($service_name, $port);

	#service can be a port (eg 80) or a name in /etc/services (eg http)
	#need to find the other version too

	#if $service contains no digits, then assume it is a name and find corresponding port 
	if ($service =~ /\D/)	
		{
		#$service is a name
		$service_name = $service;
		$port = getservbyname($service, $protocol);
		if ($port eq '')
			{
			print "service $service_name does not give a valid port number in /etc/services \n";
			exit 1;
			}
		else
			{
			#print "port for service $service_name is $port \n";
			}
		}
	else 	
		{
		#$service is a port number and not a service name
		$port = $service;
		$service_name = getservbyport($service,$protocol);
		if ($service_name eq '') 
			{
			#print "no name available for service on port $port \n";
			$service_name = $service;
			}
		else
			{
			#print "service for port $service is $service_name \n";
			}
		}
	return ($service_name, $port);
	}

sub set_monitor
	{
	#does this service have a monitor? if not set fping as the default monitor
	$monitor = "fping";
	foreach $temp (@monitors)
		{
		if ($temp eq $service_name)
			{
			$monitor = $temp;
			}
		}			
	}
		
sub write_mon_hostgroup
	{
	print MON_FILE "#------\n";
	print MON_FILE "hostgroup LVS$server_counter $server_ip \n";
	print MON_FILE "\n"; 
	print MON_FILE "watch LVS$server_counter \n";
	print MON_FILE "#the string/text following service (to OEL) is put in header of mail messages\n"; 
	#print MON_FILE "#The mon docs imply that a string is required (ie no blanks), but the test.alert and mail.alert handle text (several words) just fine.\n"; 
	#print MON_FILE "#virtualserver.alert bombs if text is sent (appears to be multiple parameters to Getopt), so the text is encased in \"\".\n"; 
	print MON_FILE "service \"$service_name on LVS$server_counter $server_ip\" \n";
	print MON_FILE "	interval 15s \n";
	if ($monitor eq "dns")
		{
		#the dns.monitor needs to know a zone and the master DNS machine for that zone.
		#the monitor checks the master for the SOA/serial number of the zone
		#and then requests the same info from the machine being monitored (here the realserver)
		#Since any zone will do, we will monitor the domain that the LVS machine belongs to.
		#Since the only domain the script knows about is the domain of the LVS, the script
		#will look for the LVS domain in /etc/resolv.conf
		#The director will be the master machine for the domain, whose only function need be to serve up
		#a serial number for the domain and the IP/name of the director machine.
		#The realservers need only be caching secondaries for the LVS domain. 
		#For real queries they will have to go out onto the net anyhow.
		#Once the LVS is running, if the files on the master machine are updated, 
		#the serial number on the copy on the realservers will not be updated till the 
		#timeout/refresh period (usually several hours).
		#In this case named on the realservers will have to be shutdown, 
		#files used as cache backup deleted and then named on the realservers restarted 
		#

		#look for domain in resolv.conf, then look for SOA nameserver for domain
		my $resolv_conf_file = "/etc/resolv.conf";
		open (RESOLV_CONF_FILE, "<$resolv_conf_file")|| die "Can't read $resolv_conf_file $!\n";
		my $domain;
		my $ns;
		my $line_counter=0;
		while (<RESOLV_CONF_FILE>)
			{
			$line_counter++;
			chomp;
			#print "resolv.conf $_\n";
			if ($_ =~ /^domain (.*) */)
				{
				$domain = $1;
				#print "domain: $domain $line_counter\n";
				} 
			}
		#find the nameservers for $domain
		#code modelled on dns.monitor (C) 1998 David Eckelkamp
		#and on Net::DNS man page
		my $res = new Net::DNS::Resolver;
		my $query = $res->query ($domain, "SOA");
		if ($query)
			{
			#($query->answer)[0] ->print;
			unless ($query->header->aa)
				{
				#print "nameserver not authoritative for $domain \n";
				}
			$ns = ($query->answer)[0]->string;
			#print "ns $ns\n";
			#non-greedy matching	
			if ($ns =~ /SOA\s+(.*?)\.\s+/)
				{
				#print "chopped ns $1 \n";
				$ns = $1;
				print MON_FILE "	monitor $monitor.monitor -zone $domain -master $ns \n";
				}
			}
		else
			{
			print "search for SOA for $domain failed: ", $res->errorstring, "\n";
			exit 1;
			}

		#search for nameservers (is SOA's)
		#my $query = $res->query($domain, "NS");
		#my $rr;
		#if ($query)
		#	{
		#	foreach $rr ($query->answer)
		#		{
		#		next unless $rr->type eq "NS";
		#		print $rr->nsdname, "\n";
		#		}
		#	}
		#else
		#	{
		#	print "search for nameserver for $domain failed: ", $res->errorstring, "\n";
		#	}
#		print MON_FILE "	monitor $monitor.monitor -zone mack.net -master di.mack.net \n";
		}
	else
		{
		print MON_FILE "	monitor $monitor.monitor \n";
		}
	print MON_FILE "	allow_empty_group \n";
	print MON_FILE "	period wd {Sun-Sat} \n";
	print MON_FILE "		alert mail.alert root \n";
	print MON_FILE "		upalert mail.alert root \n";
	print MON_FILE "		#-V is virtual service, -R is remote service, -P is protocol, -A is add/delete (t|u)\n";
	if ($lvs_type =~ /tun|dr/)
		{
		print MON_FILE "		alert virtualserver.alert -A -d -P -$proto -V $vip:$port -R $server_ip \n";
		print MON_FILE "		upalert virtualserver.alert -A -a -P -$proto -V $vip:$port -R $server_ip -T $connection_parameter -w $weight\n";
		}
	elsif ($lvs_type eq "nat")
		{
		print MON_FILE "		alert virtualserver.alert -A -d -P -$proto -V $inside_ip:$port -R $server_ip:{$server_port} \n";
		print MON_FILE "		upalert virtualserver.alert -A -a -P -$proto -V $inside_ip:$port -R ${server_ip}:${server_port} -T $connection_parameter -w $weight\n";
		}
	else	
		{
		print "error: unknown LVS_type $lvs_type \n";
		}

	print MON_FILE "\n";
	print MON_FILE "#the line above must be blank\n";
	print MON_FILE "\n";
	}

sub open_files
	{
	$rc_file = "rc.lvs_$lvs_type";
	$mon_file = "mon_lvs${lvs_type}.cf";
	open (RC_FILE, ">$rc_file")|| die "Can't write to $rc_file $!\n";
	open (MON_FILE, ">$mon_file")|| die "Can't write to $mon_file $!\n";
	}

sub print_headers
	{
	print RC_FILE "#rc.lvs_${lvs_type}\n";
	print RC_FILE "#(C) Joseph Mack 1999.\n"; 
	print RC_FILE "#This file needs a version of sh that has subroutines.\n";
	print RC_FILE "#sh on Solaris 2.7 and bash on Linux are fine.\n";
	print RC_FILE "#\n";
	print RC_FILE "#Setup for a Linux Virtual Server (LVS) running in $lvs_type mode.\n";
	print RC_FILE "#autogenerated by configure_lvs.pl from lvs_${lvs_type}.conf.\n";
	print RC_FILE "#Feel free to edit.\n";
	print RC_FILE "#This code is part of the LVS project http://www.linuxvirtualserver.org\n";
	print RC_FILE "#The rc.lvs files produced by this code runs on Solaris machines, \n";
	print RC_FILE "#but I haven't been able to test whether it actually sets up a working realserver yet\n";
	print RC_FILE "#\n";
	print RC_FILE "#----------------------------------------------------------------------------------\n";
	print RC_FILE "#\n";
	print RC_FILE "if [ `uname` = \"SunOS\" ]\n";
	print RC_FILE "then\n";
	print RC_FILE "	{\n";
	print RC_FILE "	IFCONFIG=\"/usr/sbin/ifconfig\"\n";
	print RC_FILE "	NETSTAT=\"/usr/bin/netstat\"\n";
	print RC_FILE "	ROUTE=\"/usr/sbin/route\"\n";
	print RC_FILE "	#AWK=\"/usr/bin/awk\"\n";
	print RC_FILE "	#echo \"Solaris \$IFCONFIG \$NETSTAT \$ROUTE\"\n";
	print RC_FILE "	}\n";
	print RC_FILE "fi\n";
	print RC_FILE "if [ `uname` = \"Linux\" ]\n";
	print RC_FILE "then\n";
	print RC_FILE "	{\n";
	print RC_FILE "	IFCONFIG=\"/sbin/ifconfig\"\n";
	print RC_FILE "	NETSTAT=\"/bin/netstat\"\n";
	print RC_FILE "	ROUTE=\"/sbin/route\"\n";
	print RC_FILE "	AWK=\"/usr/bin/awk\"\n";
	print RC_FILE "	#echo \"Linux \$IFCONFIG \$NETSTAT \$ROUTE\"\n";
	print RC_FILE "	}\n";
	print RC_FILE "fi\n";
	print RC_FILE " \n";

	print MON_FILE "#mon_lvs${lvs_type}.cf \n";
	print MON_FILE "\n";
	print MON_FILE "#an example mon.cf file for a group of LVS machines running in $lvs_type mode\n";
	print MON_FILE "#generated by configure_lvs.pl - Feel free to edit\n";
	print MON_FILE "#This code is part of the LVS project http://proxy.iinchina.net/~wensong/ippfvs\n";
	print MON_FILE "#for more information about LVS please see http://www.linuxvirtualserver.org\n";
	print MON_FILE " \n";
	print MON_FILE "#adapted from the mon.cf script supplied by Jim Trocki as part of the mon distribution.\n";
	print MON_FILE "#for info about mon see http://www.kernel.org/software/mon/\n";
	print MON_FILE " \n";
	print MON_FILE "#<IMPORTANT NOTE FROM Jim Trocki>\n";
	print MON_FILE "#\n";
	print MON_FILE "#NB:  hostgroup and watch entries are terminated with a blank line,\n";
	print MON_FILE "#or end of file. Don\'t forget the blank lines between them or you lose.\n";
	print MON_FILE "#\n";
	print MON_FILE "#</IMPORTANT NOTE FROM Jim Trocki>\n";
	print MON_FILE "#------------------------------------------------------------------\n";
	print MON_FILE "#\n";
	print MON_FILE "#mon config info, you probably don\'t need to change this very much\n";
	print MON_FILE "#\n";
	print MON_FILE "\n";
	print MON_FILE "alertdir   = /usr/lib/mon/alert.d \n";
	print MON_FILE "mondir     = /usr/lib/mon/mon.d \n";
	print MON_FILE "#maxprocs    = 20\n";
	print MON_FILE "histlength = 100\n";
	print MON_FILE "#delay before starting\n";
	print MON_FILE "randstart = 60s\n";
	print MON_FILE "\n";
	} 

#----------------------------------------------------------
#main:

if (@ARGV == 0)
	{
	print "usage: ./configure_lvs.pl conf_file \n";
	exit;
	}

#are we running this on a kernel version that has been tested (and, ahem, thought to work)
my $uname_r = `uname -r`;
chomp ($uname_r);
#print "uname_r $uname_r \n";
#check if this is a known $uname_r

my $uname_r_known = "N";
my $uname_r_item; 
foreach $uname_r_item (@uname_r_ok)
	{
	if ($uname_r eq $uname_r_item)
		{
		$uname_r_known = "Y";
		last;
		}
	#print "$uname_r_item \n";
	}
if ($uname_r_known eq "N")
	{
	print "warning: $0 has not been tested with kernel version $uname_r \n";
	}
	

$conf_file = shift(@ARGV);
#print "conf_file: $conf_file\n";
open (CONF_FILE, "<$conf_file")|| die "Can't read $conf_file $!\n";


while (<CONF_FILE>)
	{
	$line_counter++;
	chomp;
	#chuck comments and blank lines
	next if $_ =~ /^#|^\s*$/;
	#split on '=' or white space (the '=' in the conf file is for readability only)	
	@params = split /=|\s+/;
	#useful print for debug	
	#print "$params[0] \n";

	if ($params[0] eq "LVS_TYPE")
		{
		$lvs_type = '';
		shift @params;
		if (uc($params[0]) =~/TUN/) 
			{ 
			$lvs_type = "tun"; 
			$connection_parameter= '-i';
			}
		if (uc($params[0]) =~/NAT/) 
			{ 
			$lvs_type = "nat"; 
			$connection_parameter= '-m';
			}
		if (uc($params[0]) =~/DR/) 
			{ 
			$lvs_type = "dr"; 
			$connection_parameter= '-g';
			}
		($lvs_type =~ /tun|nat|dr/) || die "unknown LVS_TYPE $lvs_type ";
		#print "LVS_TYPE: $lvs_type \n";
		open_files;
		print_headers;
		}

	if ($params[0] eq "INITIAL_STATE")
		{
		shift @params;
		if (uc($params[0]) =~/OFF/)
			{
			$initial_state = "off";
			}
		if (uc($params[0]) =~/ON/)
			{
			$initial_state = "on";
			}
		($initial_state =~ /on|off/) || die "unknown INITIAL_STATE $params[0]";
		}

	if ($params[0] eq "VIP") 
		{
		shift @params;
		($vdevice, $vip, $vnetmask, $vbroadcast) = @params;
		print RC_FILE "\n";
		print RC_FILE "setup_director() { \n";
		print RC_FILE "	echo \"LVS director \"\n";
		print RC_FILE "	echo \"\"\n";
		print RC_FILE "	echo \"clearing ipchain rules \"\n";
		print RC_FILE "	/sbin/ipchains -F \n";
		print RC_FILE "	echo \"displaying ipchain rules (should be empty) \"\n";
		print RC_FILE "	/sbin/ipchains -L \n";
		print RC_FILE "	echo \"\"\n";
		print RC_FILE "	echo \"turn on ip forwarding\"\n";
		print RC_FILE "	echo \"1\" >/proc/sys/net/ipv4/ip_forward \n";
		if (!($lvs_type =~ /nat/))
			{
			#turn off icmp redirects for VS-DR, VS-TUN, for kernels =<2.2.12
			$uname_r =~ /^2\.2\.(\d+)/ ;
			my $kernel_minor_version = $1;
			#print "kernel minor version $kernel_minor_version \n";
			if ($kernel_minor_version < 13)
				{
				print RC_FILE "	echo \"kernel version $uname_r - turn off icmp redirects\"\n";
				print RC_FILE "	echo \"0\" >/proc/sys/net/ipv4/conf/eth0/send_redirects \n";
				}
			}
		print RC_FILE "	echo \"\"\n";
		print RC_FILE "	#setup VIP\n";
		print RC_FILE "	echo \"adding ethernet device and routing for VIP $vip \"\n";
		print RC_FILE "	\$IFCONFIG $vdevice $vip netmask $vnetmask broadcast $vbroadcast up \n";
		print RC_FILE "	\$ROUTE add -host $vip dev $vdevice \n";
		print RC_FILE "	echo \"listing ifconfig info for VIP $vip \"\n";
		print RC_FILE "	\$IFCONFIG $vdevice \n";
		print RC_FILE "	echo -n \"checking VIP $vip is reachable from director: \"\n";
		print RC_FILE "	/usr/local/bin/fping $vip \n";
		print RC_FILE "	echo \"listing routing info for VIP $vip \"\n";
		print RC_FILE "	\$NETSTAT -rn | grep $vip \n";
		print RC_FILE "	echo \"\"\n";
		print RC_FILE "\n";
		print RC_FILE "	echo \"clearing ipvsadm table \"\n";
		print RC_FILE "	/sbin/ipvsadm -C \n";
		print RC_FILE "	echo \"\"\n";
		print RC_FILE "\n";
		#print "VIP_DEVICE = $vdevice \n";
		}

	if ($params[0] eq "DIRECTOR_INSIDEIP") 
		{
		shift @params;
		($inside_device, $inside_ip, $inside_network, $inside_netmask, $inside_broadcast) = @params;
		#
		print RC_FILE "\n";
		#print RC_FILE "setup_director() { \n";
		#print RC_FILE "	echo \"LVS director \"\n";
		#print RC_FILE "	echo \"turning on masquerading \"\n";
		#print RC_FILE "	#setup masquerading \n";
		#print RC_FILE "	echo \"1\" >/proc/sys/net/ipv4/ip_forward \n";
		
		if ($lvs_type =~ /nat/)
		{
			my $classless;
			$classless = dotquad_to_classless($inside_netmask);
			print RC_FILE "";	
			print RC_FILE "	echo \"installing ipchain rules\"\n";
			print RC_FILE "	/sbin/ipchains -A forward -j MASQ -s ${inside_network}/${classless} -d 0.0.0.0/0\n";
			print RC_FILE "	echo \"ipchain rules \"\n";
			print RC_FILE "	/sbin/ipchains -L\n";
		}
		#
		print RC_FILE "\n";
		print RC_FILE "	#setup LVS_INSIDEIP\n";
		print RC_FILE "	echo \"adding ethernet device and routing for DIRECTOR_INSIDEIP $inside_ip \"\n";
		#print RC_FILE "	\$IFCONFIG $vdevice $vip netmask $vnetmask broadcast $vbroadcast up \n";
		print RC_FILE "	\$IFCONFIG $inside_device $inside_ip netmask $inside_netmask broadcast $inside_broadcast up \n";
		print RC_FILE "	\$ROUTE add -net $inside_network netmask $inside_netmask dev $inside_device \n";
		print RC_FILE "	echo -n \"checking if director INSIDEIP $inside_ip is reachable from director $vip - \"\n";
		print RC_FILE "	/usr/local/bin/fping $inside_ip \n";
		#print RC_FILE "\n";
		#print "VIP_DEVICE = $inside_device \n";
		#print RC_FILE "	echo \"clearing ipvsadm table \"\n";
		#print RC_FILE "	/sbin/ipvsadm -C \n";
		print RC_FILE "	echo \"\"\n";
		print RC_FILE "\n";
		}

	if ($params[0] eq "SERVER_LVS_DEVICE")
		{
		shift @params;
		$slvsdevice= $params[0];
		#tricky, if have got here have finished setting up the director, 
		#so close function setup_director() \n";
		print RC_FILE "\n";
		print RC_FILE "	echo \"\"\n";
		print RC_FILE "	#display current settings\n";
		print RC_FILE "	echo \"ipvsadm settings \"\n";
		print RC_FILE "	/sbin/ipvsadm \n";
		print RC_FILE "	echo \"\"\n";
		print RC_FILE "	}\n";
		print RC_FILE "\n";

		#open subroutine for server
		
		print RC_FILE "setup_realserver() {\n";
		print RC_FILE "	echo \"LVS server \"\n";
		print RC_FILE "	echo \" \"\n";

		if ($lvs_type =~ /tun|dr/)
			{
			print RC_FILE "	echo -n \"checking for VIP on server - \"\n";
			print RC_FILE "	if [ `\$IFCONFIG $slvsdevice | grep -c \"$vip \"` = \"0\" ] \n";
			print RC_FILE "	then \n";
			print RC_FILE "	{ \n";
			print RC_FILE "		echo \"not found, good.\"\n";
			print RC_FILE "		echo \"When we find the VIP, it will be on director. \"\n";
			print RC_FILE "		echo -n \"pinging VIP (hopefully on director): \"\n";		
			print RC_FILE "		/usr/local/bin/fping $vip \n";	
			print RC_FILE "		if [ \"\$\?\" != \"0\" ] \n";
			print RC_FILE "		then \n";
			print RC_FILE "		{ \n";		
			print RC_FILE "			echo \"$vip not found, presumably director has not been set up yet\"\n";
			print RC_FILE "			echo \"rerun this script after setting up the director if you want all checks to be positive\"\n";
			print RC_FILE "		} \n";
			print RC_FILE "		else \n";
			print RC_FILE "		{ \n";
			print RC_FILE "			echo \"$vip found, good. It's not on this server, assume it's on the director. \"\n";
			print RC_FILE "		} \n";
			print RC_FILE "		fi \n";
			print RC_FILE "	} \n";
			print RC_FILE "	else \n";
			print RC_FILE "	{ \n";
			print RC_FILE "		echo \"already on server. \"\n";
			print RC_FILE "		echo \"Presumably this script has been run previously on this machine. \"\n";
			print RC_FILE "		echo \"Won't be able to check connection to director. \"\n";
			print RC_FILE "		echo \"To be sure, you can delete (down) all ifconfig entries containing IP $vip and run this script again. \"\n";
			print RC_FILE "		echo \"\"\n";
			print RC_FILE "	} \n";
			print RC_FILE "	fi \n";
			print RC_FILE "	\n";
			print RC_FILE "	#install non-arping network device on Server\n";
			print RC_FILE "	echo \"installing non-arping $slvsdevice network device with IP $vip on this server\"\n";
			print RC_FILE "	\$IFCONFIG $slvsdevice $vip netmask 255.255.255.255 broadcast $vbroadcast up\n";
			print RC_FILE "	echo \"listing ifconfig info for VIP $vip \"\n";
			print RC_FILE "	\$IFCONFIG $slvsdevice \n";
			print RC_FILE "	echo \"installing route for VIP $vip \"\n";
			print RC_FILE "	\$ROUTE add -host $vip dev $slvsdevice \n";
			print RC_FILE "	echo \"listing routing info for VIP $vip \"\n";
			print RC_FILE "	\$NETSTAT -rn | grep \"$vip \"\n";
			print RC_FILE "	if [ `\$NETSTAT -rn | grep -c \"$vip \"` = \"1\" ] \n";
			print RC_FILE "	then \n";
			print RC_FILE "	{\n";
			print RC_FILE "	echo \"only one entry for $vip found, good \"\n";
			print RC_FILE "	#\$NETSTAT -rn | grep $vip \n";
			print RC_FILE "	}\n";
			print RC_FILE "	else\n";
			print RC_FILE "	{\n";
			print RC_FILE "	echo \"Error;there should only be one entry and it should contain the device $slvsdevice.\"\n";
			print RC_FILE "	echo -n \"extra entry found for $vip on device \"\n";
			print RC_FILE "	\$NETSTAT -rn | grep $vip | grep -v $slvsdevice | \$AWK \'{print \$8}\' \n";
			print RC_FILE "	echo \"presumably from other installs. \"\n";
			print RC_FILE "	echo \"removing entry for extra device. \"\n";
			print RC_FILE "	ifconfig `\$NETSTAT -rn | grep $vip | grep -v $slvsdevice | \$AWK \'{print \$8}\'` \"$vip \" down \n";
			print RC_FILE "	echo \"listing ifconfig info for VIP $vip again\"\n";
			print RC_FILE "	\$NETSTAT -rn | grep \"$vip \" \n";
			print RC_FILE "	echo \"To be sure, rerun this script again or remove extra routing entries by hand. \"\n";
			print RC_FILE "	}\n";
			print RC_FILE "	fi \n";
			print RC_FILE "	echo \"\"\n";
			print RC_FILE "	echo \"LVS Notice: make sure each server has a route to the outside world/client(s) \" \n";
			print RC_FILE "	echo \"This will normally be setup by rc.inet1 \" \n";
			print RC_FILE "\n";
			#print RC_FILE "#----------end rc.ipvs_${lvs_type}server------------------------------------\n";
			}
		elsif ($lvs_type eq "nat")
			{
			#dotquad_to_classless($inside_netmask);
			print RC_FILE "	#all packets from server go to INSIDEIP of director \n";
			print RC_FILE "	#you must make the INSIDEIP the default gw \n";
			print RC_FILE "	#just adding a route to the INSIDEUP network doesn't work\n";
			print RC_FILE "	echo \"making INSIDEIP $inside_ip the default gw for LVS packets from server (network $inside_network) \" \n";
			print RC_FILE "	\$ROUTE add default gw $inside_ip netmask 0.0.0.0 metric 1 \n";
			#print RC_FILE "	#This does not work \n";
			#print RC_FILE "	echo \"setup route for servers on network $inside_network \" \n";
			#print RC_FILE "	\$ROUTE add -net $inside_network netmask $inside_netmask dev $slvsdevice \n";
			print RC_FILE "	\$NETSTAT -rn | grep $inside_network \n";
			#print RC_FILE "#----------end rc.ipvs_${lvs_type}server------------------------------------\n";
			}
		else
			{
			print "error: unknown LVS_type $lvs_type \n";
			}
		}

	if ($params[0] eq "SERVER_NET_DEVICE")
		{
		shift @params;
		$snetdevice = $params[0]; #the ethernet device on the servers 
					#that has the regular ethernet connection 	
					#for the real IP (usually eth0, le0)
		}

	if ($params[0] eq "SERVICE")
		{
		#initialise a few things, these may be set from a previous "SERVICE" line
		my $persistence='';
		my $t_switch='';
		my $timeout='';
		my $temp; #temp could be a scheduler, ppc, or an IP
		#with kernel 2.2.13, SERVICE lines do not have to specify a scheduler if ppc is used
		#no scheduling (will use default, wlc), ppc with no timeout (will use default, 360sec)
		#SERVICE=t https ppc 192.168.1.1 
		#or
		#explicit scheduling (here wrr), pcc [timeout]
		#SERVICE=t https wrr ppc 192.168.1.1
		#
		#and/or explicit timeout (here 1800 sec)
		#SERVICE=t https ppc -t 1800 192.168.1.1
		#parsing this with perl is no fun, maybe I should use yacc/flex ;-/
		#I should rewrite this :-(
		#
		shift @params; #@param[0] is "SERVICE"
		($proto, $service, $temp) = @params; 
		shift @params; #shift out 2 parameters
		shift @params;
		#if $temp proves to be an IP don't want to shift it 
		#shift @params;

		#parse protocol
		if ($proto eq 't'){$protocol = "tcp";} 
		elsif ($proto eq 'u'){$protocol = "udp";} 
		else {print "error line $line_counter: unknown protocol $proto \n";}

		($service_name, $port) = parse_service($service);
		#don't have a VIP with VS_NAT, instead have an $inside_ip
		#(note - in other places I've made VS_NAT appear to use a VIP, 
		#I should make sure that VS_NAT uses the VIP notation everywhere)
		#if (defined $inside_ip) {$vip = $inside_ip;} 
		set_monitor;

		#handle scheduler/persistance
		#the next params will be one or both of scheduler/persistance (or neither, in which case it's an IP)
		#the first string will be one of rr|wrr|lc|wlc|pcc|ppc
		#if we got rr|wrr|lc|wlc|pcc then scheduler is set explicitely 
		#if pcc is found it can only be with kernel =<2.2.10, although I don't check the kernel version

		my $scheduler_known = "N";
		my $scheduler_item; 
		foreach $scheduler_item (@schedulers)
			{
			if ($temp eq $scheduler_item)
				{
				#print "scheduler $temp \n";
				$scheduler_known = "Y";
				#can shift out @params now
				$scheduler = $temp;
				shift @params;
				last;
				}
			#print "$scheduler_item \n";
			}
		if ($scheduler_known eq "Y")
			{
			#print "scheduler $scheduler found \n";
			}
		else
			{
			#print "scheduler not explicitely set \n";
			if ($temp eq "ppc")
				{
				#print "persistence found, set default scheduler, ";
				$persistence="ppc";
				#ppc isn't a scheduler, 
				#zero out $scheduler, so can assemble ipvsadm command below
				$scheduler = '';
				#can shift out @params now
				shift @params;
				#is timeout specified
				if ($params[0] eq "-t")
					{
					#non-default timeout, pop 2 strings 
					($t_switch, $timeout) = @params;
					shift @params; #shift out the 2 parameters
					shift @params;
					#print "persistence timeout $timeout \n";
					}
				else
					{
					#print "persistence timeout not specified, will use default (360 sec) \n";
					}
				}
			else
				{
				#old way was to exit here with an error
				#print "error: scheduler $scheduler is not one of rr|wrr|lc|wlc|pcc|pcc \n";
				#exit 1;

				#now setup default scheduler
				# $temp should be an IP
				$scheduler = $default_scheduler;
				#print "next parameter $temp is an IP, scheduler not specified, use default $scheduler\n";
				}
			}
		#have handled cases where a scheduler or persistence is specified
		#now handle case when both are specified
		#will have to look for next arguement(s)
		#if persistence is not set, it will be the IP of the realserver handling the service
		#if persistence is set, it will be "pcc [-t timeout]" 
	
		if ($params[0] eq "ppc")
			{
			#print "persistence specified, non default scheduler set to $scheduler \n";
			$persistence = "ppc";
			shift @params;
			#params[0] will either be "-t" or an IP
			#look for a -t switch specifying a (non-default) timeout
			if ($params[0] eq "-t")
				{
				#non-default timeout, pop 2 strings 
				($t_switch, $timeout) = @params;
				shift @params; #shift out the 2 parameters
				shift @params;
				#print "persistence timeout $timeout \n";
				}
			else
				{
				#print "persistence found, timeout not specified, will use default (360 sec) \n";
				}
			}
		else
			{
			#print "next param $params[0] is an IP \n";
			}

		#at last, setup connection
		print RC_FILE "\n";
		print RC_FILE "	#setup servers for $service \n";
		print RC_FILE "	#service $port uses";
		if ($scheduler eq '')
			{
			print RC_FILE " default scheduler (wlc)";
			}
		else
			{
			print RC_FILE " $scheduler scheduler";
			}
		if ($persistence eq "ppc")
			{
			print RC_FILE ", persistent connection";
			if ($timeout eq '')
				{
				print RC_FILE " with default timeout (360 sec) ";
				}
			else
				{
				print RC_FILE " with timeout $timeout sec ";
				}
			}
		print RC_FILE "\n";

		print RC_FILE "	/sbin/ipvsadm -A -$proto $vip:$port";
		if ($scheduler eq '')
			{
			#;
			}
		else
			{
			print RC_FILE " -s $scheduler";
			}
		if ($persistence eq "ppc")
			{
			print RC_FILE " -p";
			if ($timeout eq '')
				{
				#;
				}
			else
				{
				print RC_FILE " $timeout ";
				}
			}
		print RC_FILE "\n";
			

		#print RC_FILE "	/sbin/ipvsadm -A -$proto $vip:$port -p $timeout \n";
		#print RC_FILE "	/sbin/ipvsadm -A -$proto $vip:$port -s $scheduler \n";


		while ($server=$params[0])
			{
			#my ($server_port, $server_ipport, $server_service_name);
			shift @params;
			#parse server
			#server will be ip[,weight] for tun|dr
			#server will be ip:port[,weight] for nat
			($server_ipport, $weight) = split /,/, $server;
			if ($weight eq ''){$weight = 1;}
			if ($initial_state eq "off")
				{$weight_lvsadm = 0;}
			else
				{$weight_lvsadm = $weight;}
			
			($server_ip, $server_port) = split /:/, $server_ipport;
			#add IP of server to @servers
			push (@servers, $server_ip);

			#accumulate the number of servers out there
			#use later as a test to see if I'm the director or the server
			#don't count 127.0.0.1, all machines have it.
			#
			unless ($server_ip eq "127.0.0.1") {$server_counter++};
			#print "server_counter $server_counter\n";
			
			if ($lvs_type =~ /tun|dr/)
				{ 
				#print "director tun|dr \n";
				print RC_FILE "	echo \"adding service $port to realserver $server_ip using connection type $lvs_type weight $weight_lvsadm\"\n";
				print RC_FILE "	/sbin/ipvsadm -a -$proto $vip:$port -R $server_ip $connection_parameter -w $weight_lvsadm\n";
				print RC_FILE "	echo -n \"checking realserver $server_ip reachable from director $vip - \"\n";
				print RC_FILE "	/usr/local/bin/fping $server_ip \n";
				if ($initial_state eq "off")
					{
					#if start with initial state off, 
					#then load ipvsadm commands above 
					#(to test that they can be loaded at run time)	
					#then delete the specific entry again
					#this way the scheduling is still on
					#if instead did ipvsadm -C, then the scheduling would be deleted too
					print RC_FILE "	echo \"director starting in off state - clearing ipvsadm entry \"\n";
					print RC_FILE "	/sbin/ipvsadm -d -$proto $vip:$port -r ${server_ip} \n";
					print RC_FILE "	echo \"\"\n";
					}
				}
			elsif ($lvs_type eq "nat")
				{
				#for nat have to find port number for server too
				($server_service_name, $server_port) = parse_service($server_port);	
				#print "director nat\n";
				#print RC_FILE "	/sbin/ipvsadm -a -$proto $inside_ip:$port -r ${server_ip}:${server_port} $connection_parameter -w $weight_lvsadm\n";
				print RC_FILE "	echo \"adding service $port to realserver ${server_ip}\"\n";
				print RC_FILE "	/sbin/ipvsadm -a -$proto $vip:$port -r ${server_ip}:${server_port} $connection_parameter -w $weight_lvsadm\n";
				print RC_FILE "	echo -n \"checking if server reachable from director $vip: \"\n";
				print RC_FILE "	/usr/local/bin/fping $server_ip \n";
				if ($initial_state eq "off")
					{
					#if start with initial state off, 
					#then load ipvsadm commands above 
					#(to test that they can be loaded at run time)	
					#then delete the specific entry again
					#this way the scheduling is still on
					#if instead did ipvsadm -C, then the scheduling would be deleted too
					print RC_FILE "	echo \"director starting in off state - clearing ipvsadm ientry \"\n";
					print RC_FILE "	/sbin/ipvsadm -d -$proto $vip:$port -r ${server_ip}:${server_port} \n";
					print RC_FILE "	echo \"\"\n";
					}
				}
			else
				{
				print "error: unknown LVS_type $lvs_type \n";
				}


			write_mon_hostgroup;

			}
		#print RC_FILE "\n";
		}
	}

#close setup_realserver();
print RC_FILE "} \n";
print RC_FILE "\n";

print RC_FILE "#Determine if I'm a realserver, director (or fail)\n";
print RC_FILE "#Am I a server?\n";
print RC_FILE "#The IPs listed by ifconfig are grep'ed against the list of server IPs.\n";
print RC_FILE "#If there are no matches (== 0), then I'm not a server.\n";
print RC_FILE "#(Note the blank after the IP.\n"; 
print RC_FILE "#Without the blank, \"192.168.1.1\" matches \"192.168.1.1x\", but \"192.168.1.1 \" will not.\n";
print RC_FILE "#a similar problem exists at the front of the IP, but there's no easy way to handle it.\n";
print RC_FILE "#In linux the previous char is a blank or \'\', in solaris it's \'#\' \n";
print RC_FILE "#If I'm not a server, then maybe I'm the director. If I'm the director, then ipvsadm will be on this machine.\n";
print RC_FILE "#If this fails, give up, exit and cry for help\n";

print RC_FILE "\n";
print RC_FILE "if \\\n";
while ($server_ip = $servers[0])
	{
	#loop through IPs of known realservers comparing them to the IP of this machine.
	shift @servers;
	#don't test for 127.0.0.1, all machines have this, and they weren't counted with $server_counter
	if ($server_ip eq "127.0.0.1") {next;}
	$server_counter--;
	#print "server_counter $server_counter $server_ip\n";
	#print RC_FILE "[ `\$IFCONFIG $snetdevice | grep -c \"[ #]$server_ip \"` = \"0\" ] ";
	print RC_FILE "[ `\$IFCONFIG $snetdevice | grep -c \"$server_ip \"` = \"0\" ] ";
	#are we at the last server?
	if ($server_counter > 0) 
		{
		#not last server
		print RC_FILE " && \\\n";
		}
#	else	
#		{
#		#last server
#		print RC_FILE " \n";
#		}
	}
#print RC_FILE "if [ `\$IFCONFIG $snetdevice | grep -c \"192.168.1.8 \"` = \"0\" ] \n";
print RC_FILE "\n";
print RC_FILE "then \n";
print RC_FILE "	{\n";
print RC_FILE "	#I'm the director, then /sbin/ipvsadm should be on this machine\n";
print RC_FILE "	if [ -x /sbin/ipvsadm ] \n";
print RC_FILE "	then\n";
print RC_FILE "		{\n";
print RC_FILE "		setup_director\n";
print RC_FILE "		}\n";
print RC_FILE "	else\n";
print RC_FILE "		{\n";
print RC_FILE "		echo \"error: not a server, not a director\"\n";
print RC_FILE "		echo \"I'm not a server machine - this machine doesn\'t have the IP of any of the servers\"\n";
print RC_FILE "		echo \"I'm not a director - I can't execute /sbin/ipvsadm\"\n";
print RC_FILE "		echo \"bummer - exiting.\"\n";
print RC_FILE "		#return error in case this script is executed by another script\n";
print RC_FILE "		return 1\n";
print RC_FILE "		}\n";
print RC_FILE "	fi\n";
print RC_FILE "	}\n";
print RC_FILE "else \n";
print RC_FILE "	{\n";
print RC_FILE "	setup_realserver\n";
print RC_FILE "	}\n";
print RC_FILE "fi\n";
print RC_FILE "\n";

print RC_FILE "#----rc.lvs_${lvs_type}-------------------------------------------------------------\n";
print MON_FILE      "#----mon_lvs${lvs_type}.cf------------------------------------------------------------------\n";
#-----configure_lvs-2.2.pl---------------------------------


